#!/usr/bin/env perl

use strict;
use Getopt::Long;
use StructureParse;

sub main{

	# Options handling
	my $entpdb;
	my $opt_help = (@ARGV) ? 0 : 1;
	my $fileXML;
	my $uniq_intp_file;
	my $blind = 0;
	my $N = 50;
	my $pId = 0.5;
	my $digits = 4;
	my $Zoption;
	
	# Residual variables in this v1.1
	my ($all_atoms, $any_atom, $any_location) = (undef) x 3;
	my $ATM_or_CA = ($all_atoms) ? 'all_atm' : 'alpha_C';

	GetOptions(
	'help' => \$opt_help,
	'input=s' => \$entpdb,
	'xml=s' => \$fileXML,
	'energy=s' => \$uniq_intp_file,
	'probability=s' => \$pId,
	'blind' => \$blind,
	'digits=s' => \$digits,
	'number=s' => \$N,
	'zscore' => \$Zoption);
	display_help() if ($opt_help);

	# Exception handling
	unless ($opt_help){
		die "\nError: -i argument required\n" if (!$entpdb);
		die "\nError: -x argument required\n" if (!$fileXML);
		die "\nError: -e argument required\n" if (!$uniq_intp_file);
	}
	die "\nError: -N argument must be >2\n" if ($N <= 2);
	die "\nError: -p argument must belong to ]0;1[\n" if ($pId <= 0 or $pId >= 1);
	die "\nError: -d argument must be >0\n" if ($digits <= 0);
	die "\nError: *.xml file expected for -x argument\n" if ($fileXML !~ m/\.xml$/);
	print STDERR "\nWarning: unexpected file extension for -i argument\n" if ($entpdb !~ m/\.pdb|\.ent/);
	
	my %empty_hash_coord; # will contain atomic coordinates of query protein residues
	my %hash_intp = intp_to_hash($uniq_intp_file); # contains the pseudo-energy profiles

	# testing first and last expected keys of %hash_intp
	unless (exists $hash_intp{'0.05_0.15_I_I_A_A'} && exists $hash_intp{'14.85_14.95_I_I_Y_Y'}){
		die "\nError: corrupted scoring table for -e argument\n"
	}
	
	print "\nFile: $entpdb\n\n"; # the query protein structure file
	
	my @pdb_file = read_file($entpdb);
	my ($start, $end) = find_best_model(\@pdb_file); # if several models in the file (shouldn't happen)
	my @chainList = chains_check_pdb($start, $end, \@pdb_file);
	
	# Consistency check between the structure file and the PDBTM XML file
	my @xml_file;
	unless ($any_location){
		@xml_file = read_file($fileXML);
		my @xmlChains = chains_check_xml(\@xml_file);
		@chainList = compare_chains(\@chainList, \@xmlChains, $entpdb);
	}
	
	# Processing multimeric protein structures requires to assign new numbers to protein atoms
	# get_coordinates() fills %empty_hash_coord and returns the (new) number of the last atom that
	# have been processed
	my $last_std_num = 0;
	foreach my $chain (@chainList){
		my $first_num = chain_first_residue($start, \@pdb_file, $chain); 
		my @residues_I = ($any_location) ? () : intern_residues($chain, \@xml_file);
		$last_std_num = get_coordinates(\%empty_hash_coord, \@pdb_file,
		                                $chain, \@residues_I, $start, $end,
		                                $first_num, $last_std_num, $entpdb,
		                                $ATM_or_CA, $any_location, $any_atom);
	}
	
	# Size of %empty_hash_coord = number of membrane residues
	my $length = 0;
	while (each %empty_hash_coord){
		$length++;
	}
	print "\n$length residues in the lipid bilayer\n\n";
	
	# Pseudo-energy computation
	my $score = calcul_score(\%empty_hash_coord, \%hash_intp, $ATM_or_CA, $any_atom);
	print "E = ";
	printf("%.${digits}f", $score);
	print "\n\n";

	# Z-score computation
	if ($Zoption){
		my @zscore_results = z_score(\%empty_hash_coord, \%hash_intp, $N, $length,
		                             $score, $pId, $ATM_or_CA, $any_atom, $blind, $digits);
	}
}


sub intp_to_hash{
# Creates a hash that contains the pseudo-energy profiles
#
# Arguments:
#     the file that contains pseudo-energy profiles
#
# Returns:
#     a hash that will be used as a scoring function

	my %hash_intp;
	my $file = shift;

	open (my $fh, '<', $file) or die "\nError while reading $file:\n\t$!\n";

	while(<$fh>){
		chomp $_;
		my ($d1,$d2,$mb1,$mb2,$aa1,$aa2,$m,$p) = split(' ',$_);
		$hash_intp{"$d1\_$d2\_$mb1\_$mb2\_$aa1\_$aa2"} = "$m $p";
	}

	return %hash_intp;
}


sub calcul_score{
# Computes the pseudo-energy of the query protein structure
#
# Arguments:
#     the reference of the hash that contains atomic coordinates of the query structure
#     the reference of the hash that contains the pseudo-energy profiles
#
# Returns:
#     the pseudo-energy of the query protein structure

	my $E = 0;
	my ($hash_coord_ref, $hash_intp_ref, $ATM_or_CA, $any_atom) = splice(@_, 0, 4);
	my %hash_intp = %$hash_intp_ref;

	my @the20AAs = all_atoms_or_CA($ATM_or_CA, $any_atom);
	my $factor = ($ATM_or_CA eq 'alpha_C') ? 1 : 10E5;

	# Reminder: %hash_coord keys are the residues (new) numbers
	my @res_num = sort {$a<=>$b}(keys(%{$hash_coord_ref}));

	for my $i (0..$#res_num){
		for ($i..$#res_num){
			if ($res_num[$_] > $res_num[$i]+4*$factor){
				my $distance = sqrt(
                                    (${$hash_coord_ref}{$res_num[$i]}[1] -
                                     ${$hash_coord_ref}{$res_num[$_]}[1]) ** 2 +
                                    (${$hash_coord_ref}{$res_num[$i]}[2] -
                                     ${$hash_coord_ref}{$res_num[$_]}[2]) ** 2 +
                                    (${$hash_coord_ref}{$res_num[$i]}[3] -
                                     ${$hash_coord_ref}{$res_num[$_]}[3]) ** 2
				);

				if ($distance < 15 and $distance > 0){
					my $AA1 = ${$hash_coord_ref}{$res_num[$i]}[0];
					my $AA2 = ${$hash_coord_ref}{$res_num[$_]}[0];
					my $state1 = ${$hash_coord_ref}{$res_num[$i]}[4];
					my $state2 = ${$hash_coord_ref}{$res_num[$_]}[4];

					# interpolation
					$distance = 14.95 if ($distance > 14.95);

					my $limit = $distance * 10;
					$limit = int($limit);
					$limit = ($limit + 0.5)/10;

					my ($distance_class_inf, $distance_class_sup) = ($limit) x 2;
					$distance_class_sup = $limit + 0.1 if ($distance > $limit);
					$distance_class_inf = $limit - 0.1 if ($distance <= $limit);
					my $key_distances = "$distance_class_inf\_$distance_class_sup";

					# after the distance is calculated, it has to be
					# pushed in the array that corresponds to the proper key

					# position in the @the20AAs array for residues 1 and 2
					# residue 2 position in the @the20AAs array
					my $naa1 = residue_position($AA1, \@the20AAs);
					my $naa2 = residue_position($AA2, \@the20AAs);

					my $key_states_AAs;

					# if AA1 = CYS and AA2 = ALA, with this condition,
					# "ALA CYS" is pushed and not "CYS ALA"
					if ($AA1 eq $AA2 and $state1 ne $state2){
						$key_states_AAs = ($state1 eq "I") ? "_$state1\_$state2\_$AA1\_$AA2" :
						                                     "_$state2\_$state1\_$AA2\_$AA1";
					}

					else{
						$key_states_AAs = ($naa1 <= $naa2) ? "_$state1\_$state2\_$AA1\_$AA2" :
						                                     "_$state2\_$state1\_$AA2\_$AA1";
					}

					my $key = $key_distances.$key_states_AAs;

					(exists $hash_intp{$key}) ? $E = calcul_energy($hash_intp{$key}, $distance, $E) :
					                            print STDERR "Warning! Unexpected residue or atom: $key\n";
				}
			}
		}
	}

	return $E;
}


sub calcul_energy{
	my $mp = shift;
	my $distance = shift;
	my $E = shift;

	my ($m, $p) = split (' ', $mp);
	$E+=($m * $distance)+$p;

	return $E;
}


sub z_score{
# Generates N sequence decoys by random permutations of residue positions in the query structure
# Computes N pseudo-energies from the generated random sequence decoys
# Computes the pseudo-enery Z-score of the query protein structure
#
# Arguments:
#     reference of the hash that contains atomic coordinates of the query structure
#     reference of the hash that contains the pseudo-energy profiles
#     number of random sequence decoys to generate
#     number of residues in the lipid bilayer ($length)
#     pseudo-energy of the query structure
#     sequence identity between the generated decoys and the query structure
#
# Displays:
#     computation progress
#     pseudo-enery Z-score of the query structure

	my ($ref_empty_hash_coord, $ref_hash_intp, $N, $length,
	    $score, $pId, $ATM_or_CA, $any_atom, $blind, $digits) = splice(@_, 0, 10);

	my %hash_intp = %$ref_hash_intp;
	my %empty_hash_coord = %$ref_empty_hash_coord;

	my ($progress, $prev_prog) = (0) x 2;

	# Transforms %empty_hash_coord into an array (of array)
	# -> to perform the permutations
	my @array_coord = map {$empty_hash_coord{$_}} sort {$a<=>$b}(keys(%empty_hash_coord));

	# Determining the original sequence
	# Attention : here, the chain name is already known
	my @initial_sequence = map {${$_}[0]} @array_coord;

	# Generating 2 random numbers
	my ($randNum1, $randNum2);
	my $range = $#array_coord;
	my @random_scores = ();
	my $n = 1;

	while ($n <= $N){
		my @shuffled_array_coord = map {[@$_]} @array_coord;

		# Permutations go on until sequence identity reaches the threshold
		my $similarity = 1; # Initial sequence identity = 100%
		while ($similarity > $pId){
			$randNum1 = int(rand($range));
			$randNum2 = int(rand($range));

			# Array modification: one (random) permutation
			my $aa1 = $shuffled_array_coord[$randNum1][0];
			my $aa2 = $shuffled_array_coord[$randNum2][0];
			$shuffled_array_coord[$randNum2][0] = $aa1;
			$shuffled_array_coord[$randNum1][0] = $aa2;

			# Determines the new sequence from the modified array
			my @new_sequence = map {${$_}[0]} @shuffled_array_coord;

			# Comparison with the former sequence
			my $counter = 0;
			for (0..$#initial_sequence){
				$counter++ if ($initial_sequence[$_] eq $new_sequence[$_]);
			}
			$similarity = $counter/$length; # Sequence identity
		}

		my %shuffled_hash_coord;
		for (0..$#shuffled_array_coord){
			my $key = $shuffled_array_coord[$_][6];
			$shuffled_hash_coord{$key} = $shuffled_array_coord[$_];
		}

		my $random_score = calcul_score(\%shuffled_hash_coord, \%hash_intp, $ATM_or_CA, $any_atom);

		# Progress display
		unless ($blind){
			$progress = $n/$N;
			my $leap = $progress-$prev_prog;

			local $| = 1;
			if ($leap >= 0.007){
				my $percent_prog = $progress*100;
				print " ";
				printf('%.0f', $percent_prog);
				print "% [random sequence decoys: $n/$N generated]";
				$prev_prog = $progress;
				print "\r";
			}
		}

		push (@random_scores, $random_score);
		$n++;
	}

	# Statistics
	my ($sum, $sumSq) = (0) x 2; # sum and sum of squares

	foreach my $value (@random_scores){
		$sum = $sum + $value;
		$sumSq = $sumSq + ($value**2);
	}

	my $mean = $sum/$N;
	my $variance = ($sumSq/$N)-($mean**2);
	my $sd = sqrt($variance);
	my $z_score = ($score-$mean)/$sd;
	print "\n\nZ = ";
	printf("%.${digits}f", $z_score);
	print "\n\n";
}


sub display_help{

	print
	"\n\tMAIDEN v1.2: Model quality Assessment for Intramembrane Domains using an ENergy criterion\n\n\n",

	"\tUsage:\n",
	"\tMAIDEN -i COORD_FILE -x PDBTM_XML -e SCORING_TABLE [-d digits] [-z] [-n num_of_decoys] [-p decoys_seqId] [-b]\n\n",

	"\tOptions:\n",
	"\t-i, --input		the protein structure file\n",
	"\t-x, --xml		the corresponding PDBTM XML file from http://pdbtm.enzim.hu/ (Tusnady et al.)\n",
	"\t-e, --energy		the pseudo-energy potential (*.intp file)\n",
	"\t-d, --digits		number of significant figures (default = 4)\n",
	"\t-z, --zscore		compute Z-score\n",
	"\t-n, --number		number of random sequence decoys for the Z-score calculation (default = 50)\n",
	"\t-p, --probability	sequence identity after permutations in the random decoys (default = 0.5)\n",
	"\t-b, --blind		turn off the display of Z-score calculations progress\n\n",

	"\tExample:\n",
	"\t./MAIDEN -i ../test/3tds.pdb -x ../test/3tds.xml -e ../src/MAIDEN.intp\n\n",
	"\tfor Windows:\n",
	"\t.\\winMAIDEN.exe -i ..\\test\\3tds.pdb -x ..\\test\\3tds.xml -e ..\\src\\MAIDEN.intp\n\n";
	
	exit 1;
}

main();

__END__

© Guillaume Postic, 2014

guillaume.postic@univ-paris-diderot.fr

This software is a computer program whose purpose is to assess the
quality of membrane protein tridimensional structures by calculating
pseudo-energies. It is a Perl program compatible with Unix-like OS.
It requires the 'StructureParse.pm' module. The 'Getopt::Long' module
must be installed.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

